import abc
from collections import OrderedDict
import numpy as np
from scipy.optimize import curve_fit as scipy_fit
from scipy.signal import find_peaks_cwt
from uncertainties import ufloat
from pruby.strategies.base import BaseStrategy, BaseStrategies
from pruby.utility import gaussian, pseudovoigt
from pruby.spectrum import Curve
from pruby.constants import R1_0, R2_0


class PeakfittingStrategy(BaseStrategy, abc.ABC):
    @abc.abstractmethod
    def peakfit(self, calc):
        raise NotImplementedError


class PeakfittingStrategies(BaseStrategies):
    registry = OrderedDict()
    strategy_type = PeakfittingStrategy


class BasePeakfittingStrategy(PeakfittingStrategy):
    @abc.abstractmethod
    def _prepare_peakfit(self, calc):
        pass

    @abc.abstractmethod
    def _assign_peaks(self, calc):
        pass

    def peakfit(self, calc):
        self._prepare_peakfit(calc)
        x = calc.peak_spectrum.focused.x
        y = calc.peak_spectrum.focused.y
        si = calc.peak_spectrum.focused.si
        calc.peak_spectrum.curve.args, pcov = \
            scipy_fit(calc.peak_spectrum.curve, xdata=x, ydata=y,
                      p0=calc.peak_spectrum.curve.args, sigma=si)
        calc.peak_spectrum.curve.uncs = np.sqrt(np.diag(pcov))
        self._assign_peaks(calc)

    @staticmethod
    def find_initial_peaks(spectrum):
        x_span = max(spectrum.x) - min(spectrum.x)
        caret_width = int(0.5 / (x_span / len(spectrum.x))) + 1
        peak_indices = find_peaks_cwt(spectrum.y, [caret_width] * 3)
        peaks = np.array([(spectrum.x[i], spectrum.y[i])
                          for i in peak_indices])
        peaks = peaks[peaks[:, 1].argsort()[::-1]]
        if len(peaks) == 0:
            r1x = R1_0.nominal_value
            r1y = max(spectrum.y)
            r2x = R2_0.nominal_value
            r2y = max(spectrum.y) / 2
        elif len(peaks) == 1:
            r1x = peaks[0][0]
            r1y = peaks[0][1]
            r2x = r1x * (1 - 0.002)
            r2y = r1y / 2
        else:
            if peaks[0][0] - 3.0 < peaks[1][0] < peaks[0][0] - 1.0:
                r1x = peaks[0][0]
                r1y = peaks[0][1]
                r2x = peaks[1][0]
                r2y = peaks[1][1]
            else:
                r1x = peaks[0][0]
                r1y = peaks[0][1]
                r2x = r1x * (1 - 0.002)
                r2y = r1y / 2
        return r1x, r1y, r2x, r2y

    @staticmethod
    def ufloat_from_curve_args(curve, index):
        return ufloat(curve.args[index], curve.uncs[index])


@PeakfittingStrategies.register(default=True)
class GaussianPeakfittingStrategy(BasePeakfittingStrategy):
    name = 'Gaussian'

    def _prepare_peakfit(self, calc):
        def two_gaussians(x, _a1, _mu1, _si1, _a2, _mu2, _si2):
            return gaussian(_a1, _mu1, _si1)(x) + gaussian(_a2, _mu2, _si2)(x)
        mu1, a1, mu2, a2 = self.find_initial_peaks(calc.peak_spectrum)
        si1 = si2 = 0.3
        calc.peak_spectrum.curve = Curve(func=two_gaussians,
                                         args=(a1, mu1, si1, a2, mu2, si2))
        peaks = self.find_initial_peaks(calc.peak_spectrum)
        calc.peak_spectrum.focus_on_points((peaks[0], peaks[2]), width=0.5)
        calc.peak_spectrum.sigma_type = 'equal'

    def _assign_peaks(self, calc):
        curve = calc.peak_spectrum.curve
        calc.r1 = self.ufloat_from_curve_args(curve, index=1)
        calc.r2 = self.ufloat_from_curve_args(curve, index=4)


@PeakfittingStrategies.register()
class PseudovoigtPeakfittingStrategy(BasePeakfittingStrategy):
    name = 'Pseudovoigt'

    def _prepare_peakfit(self, calc):
        def two_pseudovoigts(x, _a1, _mu1, _w1, _et1, _a2, _mu2, _w2, _et2):
            return pseudovoigt(_a1, _mu1, _w1, _et1)(x) + \
                   pseudovoigt(_a2, _mu2, _w2, _et2)(x)

        mu1, a1, mu2, a2 = self.find_initial_peaks(calc.peak_spectrum)
        w1 = w2 = 0.6
        et1 = et2 = 0.5
        calc.peak_spectrum.curve = Curve(
            func=two_pseudovoigts,
            args=(a1, mu1, w1, et1, a2, mu2, w2, et2))
        peaks = self.find_initial_peaks(calc.peak_spectrum)
        calc.peak_spectrum.focus_on_points((peaks[0], peaks[2]), width=1.0)
        calc.peak_spectrum.sigma_type = 'equal'

    def _assign_peaks(self, calc):
        curve = calc.peak_spectrum.curve
        calc.r1 = self.ufloat_from_curve_args(curve, index=1)
        calc.r2 = self.ufloat_from_curve_args(curve, index=5)


@PeakfittingStrategies.register()
class CamelPeakfittingStrategy(BasePeakfittingStrategy):
    name = 'Camel'

    def _prepare_peakfit(self, calc):
        def camel(x, _a1, _mu1, _si1, _a2, _mu2, _si2, _a, _si):
            return gaussian(_a2, _mu2, _si2)(x) + gaussian(_a1, _mu1, _si1)(x) \
                   + gaussian(_a, (_mu2 + _mu1) / 2, _si)(x)
        mu1, a1, mu2, a2 = self.find_initial_peaks(calc.peak_spectrum)
        si1, si2, si, a = 0.35, 0.35, 1.0, a1 / 10
        calc.peak_spectrum.curve = Curve(
            func=camel,
            args=(a1, mu1, si1, a2, mu2, si2, a, si))
        peaks = self.find_initial_peaks(calc.peak_spectrum)
        calc.peak_spectrum.focus_on_points((peaks[0], peaks[2]), width=1.0)
        calc.peak_spectrum.sigma_type = 'equal'

    def _assign_peaks(self, calc):
        curve = calc.peak_spectrum.curve
        calc.r1 = self.ufloat_from_curve_args(curve, index=1)
        calc.r2 = self.ufloat_from_curve_args(curve, index=4)


@PeakfittingStrategies.register()
class NullPeakfittingStrategy(PeakfittingStrategy):
    name = 'No peak fitting'

    def peakfit(self, calc):
        curve = Curve()



import abc
from collections import OrderedDict
from pruby.strategies.base import BaseStrategy, BaseStrategies
from pruby.utility import polynomial
from pruby.constants import T_0, UZERO


def to_wavelength(wavenumber):
    """Recalculate wavenumber in cm^-1 to wavelength in nm."""
    return 1e7 / wavenumber


def ragan_r1_position(t):
    return to_wavelength(polynomial(14423, 4.49e-2, -4.81e-4, 3.71e-7)(t))


def ragan_r2_position(t):
    return to_wavelength(polynomial(14452, 3.00e-2, -3.88e-4, 2.55e-7)(t))


def vos_r1_shift(t):
    return 0.1 * polynomial(0.0, 6.591e-2, 7.624e-5, -1.733e-7)(t - 300.0)


def vos_r2_shift(t):
    return 0.1 * polynomial(0.0, 6.554e-2, 8.670e-5, -1.099e-7)(t - 300.0)


class CorrectingStrategy(BaseStrategy, abc.ABC):
    @abc.abstractmethod
    def correct(self, calc):
        raise NotImplementedError


class CorrectingStrategies(BaseStrategies):
    registry = OrderedDict()
    strategy_type = CorrectingStrategy


@CorrectingStrategies.register(default=True)
class VosR1CorrectingStrategy(CorrectingStrategy):
    name = 'Vos R1'  # (1991)
    year = 1991
    reference = r'https://doi.org/10.1063/1.348903'

    def correct(self, calc):
        calc.t_correction = -vos_r1_shift(calc.t)


class VosR2CorrectingStrategy(CorrectingStrategy):
    """Based on doi:10.1063/1.348903"""
    name = 'Vos R2'
    year = 1991
    reference = r'https://doi.org/10.1063/1.348903'

    def correct(self, calc):
        calc.t_correction = -vos_r2_shift(calc.t)


class VosR12CorrectingStrategy(CorrectingStrategy):
    name = 'Vos average'
    year = 1991
    reference = r'https://doi.org/10.1063/1.348903'

    def correct(self, calc):
        calc.t_correction = \
            - 0.5 * vos_r2_shift(calc.t) - 0.5 * vos_r1_shift(calc.t)


@CorrectingStrategies.register()
class RaganR1CorrectingStrategy(CorrectingStrategy):
    name = 'Ragan R1'
    year = 1992
    reference = r'https://doi.org/10.1063/1.351951'

    def correct(self, calc):
        calc.t_correction = ragan_r1_position(T_0) - ragan_r1_position(calc.t)


class RaganR2CorrectingStrategy(CorrectingStrategy):
    name = 'Ragan R2'  # (1992)
    year = 1992
    reference = r'https://doi.org/10.1063/1.351951'

    def correct(self, calc):
        calc.t_correction = ragan_r2_position(T_0) - ragan_r2_position(calc.t)


class RaganR12CorrectingStrategy(CorrectingStrategy):
    name = 'Ragan average'  # (1992)
    year = 1992
    reference = r'https://doi.org/10.1063/1.351951'

    def correct(self, calc):
        calc.t_correction = (ragan_r1_position(T_0)-ragan_r1_position(calc.t) +
                             ragan_r2_position(T_0)-ragan_r2_position(calc.t))/2


@CorrectingStrategies.register()
class NoneCorrectingStrategy(CorrectingStrategy):
    name = 'None'

    def correct(self, calc):
        calc.t_correction = UZERO



import numpy as np


def polynomial(*coefficients):
    return lambda x: sum([c * x ** i for i, c in enumerate(coefficients)])


def gaussian(a, mu, si):
    return lambda x: a * np.exp(-(x - mu) ** 2 / (2. * si ** 2))


def lorentzian(a, mu, ga):
    return lambda x: (a * ga ** 2) / ((x - mu) ** 2 + ga ** 2)


def pseudovoigt(a, mu, w, et):
    si = w / (np.sqrt(8 * np.log(2)))
    ga = w / 2
    return lambda x: et * gaussian(a, mu, si)(x) + \
                     (1 - et) * lorentzian(a, mu, ga)(x)



import abc
from collections import OrderedDict
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from pruby.strategies import BaseStrategy, BaseStrategies
from pruby.utility import cycle

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pruby import PressureCalculator


mpl.use('Agg')


class DrawingStrategy(BaseStrategy, abc.ABC):
    calc: 'PressureCalculator'

    @abc.abstractmethod
    def draw(self, calc):
        raise NotImplementedError


class DrawingStrategies(BaseStrategies):
    registry = OrderedDict()
    strategy_type = DrawingStrategy


class BaseDrawingStrategy(DrawingStrategy, abc.ABC):
    def __init__(self):
        plt.close()
        mpl.use('Agg')
        self.calc: 'PressureCalculator'
        self.fig: plt.Figure = plt.figure(figsize=(8, 6), dpi=100)
        self.ax: plt.Axes = self.fig.add_subplot()
        self.color = '#000000'
        self.color_cycle = cycle(
            ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
             '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'])

    @property
    def interactive(self):
        return not bool(self.calc.output_path)

    def draw_initialize(self, calc):
        self.calc = calc
        if self.interactive:
            mpl.use('TkAgg')
        if not self.calc.fig.axes:
            self.draw_new_figure()
        elif not plt.fignum_exists(self.calc.fig.number):
            self.draw_new_figure()
        self.color = next(self.color_cycle)

    def draw_new_figure(self):
        self.fig = plt.figure(figsize=(8, 6), dpi=100)
        self.ax = self.fig.add_subplot()

    def draw_grid_and_tics(self):
        self.ax.minorticks_on()
        self.ax.grid(b=True, which='major', color='gray', alpha=0.2)
        self.ax.grid(b=True, axis='x', which='minor', color='gray', alpha=0.2)
        self.ax.grid(b=True, axis='y', which='major', color='gray', alpha=0.2)
        self.ax.tick_params(axis='x', which='minor', bottom=True)

    def draw_spectrum(self, x, y):
        if self.interactive:
            label = self.calc.dat_path
        else:
            label = f'p = {self.calc.p} GPa'
        self.ax.plot(x, y, self.color, marker='.', ls='None', label=label)

    def draw_curve(self, spectrum, bg=False):
        y = -spectrum.f if bg else spectrum.f
        ls = '--' if bg else '-'
        self.ax.plot(spectrum.x, y, self.color, ls=ls)
        self.ax.fill_between(x=spectrum.x, y1=y, color=self.color, alpha=0.2,
                             where=[_ in spectrum.focus for _ in spectrum.x])

    def draw_vline(self, x):
        self.ax.axvline(x, color=self.color)

    def draw_finalize(self):
        self.ax.set_xlim(min(self.calc.peak_spectrum.x),
                         max(self.calc.peak_spectrum.x))
        self.ax.annotate('nm', xy=(1, 0), ha='left', va='top',
                         xytext=(10, - 3 - mpl.rcParams['xtick.major.pad']),
                         xycoords='axes fraction', textcoords='offset points')
        self.ax.legend()
        self.calc.fig = self.fig
        if self.interactive:
            self.fig.show()
        else:
            self.fig.savefig(self.calc.output_path)


@DrawingStrategies.register()
class SimpleDrawingStrategy(BaseDrawingStrategy):
    name = 'Simple'

    def draw(self, calc):
        self.draw_initialize(calc)
        self.draw_grid_and_tics()
        self.draw_spectrum(calc.peak_spectrum.x, calc.peak_spectrum.y)
        self.draw_vline(calc.r1.n)
        self.draw_finalize()


@DrawingStrategies.register(default=True)
class ComplexDrawingStrategy(BaseDrawingStrategy):
    name = 'Complex'

    def draw(self, calc):
        self.draw_initialize(calc)
        self.draw_grid_and_tics()
        self.draw_spectrum(calc.peak_spectrum.x, calc.peak_spectrum.y)
        self.draw_curve(calc.peak_spectrum)
        self.draw_curve(calc.back_spectrum, bg=True)
        self.draw_vline(calc.r1.n)
        self.draw_vline(calc.r2.n)
        self.draw_finalize()


@DrawingStrategies.register()
class SingleLineDrawingStrategy(BaseDrawingStrategy):
    name = 'Single line'

    def draw_finalize(self):
        self.ax.annotate('nm', xy=(1, 0), ha='left', va='top',
                         xytext=(10, - 3 - mpl.rcParams['xtick.major.pad']),
                         xycoords='axes fraction', textcoords='offset points')
        self.ax.legend()
        self.calc.fig = self.fig
        if self.interactive:
            self.fig.show()
        else:
            self.fig.savefig(self.calc.output_path)

    def draw(self, calc):
        self.draw_initialize(calc)
        self.draw_grid_and_tics()
        self.draw_spectrum([np.NaN], [np.NaN])
        self.draw_vline(calc.r1.n)
        self.draw_finalize()



import os
import tkinter as tk
from natsort import natsorted
from pruby.constants import R1_0, T_0, P_0
from pruby.calculator import PressureCalculator
from pruby.gui.gridable import FilenameEntry, UfloatEntry, StatusBar
from pruby.gui.popups import open_file_dialogue, show_about
from pruby.strategies import \
    ReadingStrategies, \
    BackfittingStrategies, \
    PeakfittingStrategies, \
    CorrectingStrategies, \
    TranslatingStrategies, \
    DrawingStrategies


class Application(tk.Frame):
    def __init__(self, root, *args, **kwargs):
        tk.Frame.__init__(self, root, *args, **kwargs)
        self.root = root
        os.chdir(os.path.expanduser('~/'))

        # SETTING CONSTANTS
        self.r1_ref = R1_0
        self.t_ref = T_0
        self.p_ref = P_0
        self.autodraw = tk.BooleanVar(value=False)
        self.calc = PressureCalculator()
        self.ref = PressureCalculator()

        # method string variables
        self.reading_strategy = tk.StringVar(
            value=self.calc.engine.reader.name)
        self.backfitting_strategy = tk.StringVar(
            value=self.calc.engine.backfitter.name)
        self.peakfitting_strategy = tk.StringVar(
            value=self.calc.engine.peakfitter.name)
        self.correcting_strategy = tk.StringVar(
            value=self.calc.engine.corrector.name)
        self.translating_strategy = tk.StringVar(
            value=self.calc.engine.translator.name)
        self.drawing_strategy = tk.StringVar(
            value=self.calc.engine.drawer.name)

        # BUILDING TOP MENU
        self.menu = tk.Menu(self)
        self.root.config(menu=self.menu)

        self.menu_data = tk.Menu(self.menu, tearoff=0)
        self.menu.add_cascade(label="Data", menu=self.menu_data)
        self.menu_data.add_command(label='Load',
                                   command=self.open_file_dialogue)
        self.menu_data.add_command(label='Draw',
                                   command=self.data_draw)
        self.menu_data.add_command(label='To reference',
                                   command=self.save_reference)
        self.menu_data.add_command(label='From reference',
                                   command=self.load_reference)
        self.menu_data.add_checkbutton(label='Auto draw', onvalue=True,
                                       offvalue=False, variable=self.autodraw)

        self.menu_options = tk.Menu(self.menu, tearoff=0)
        self.menu.add_cascade(label="Methods", menu=self.menu_options)

        def make_options_submenu(strategy_list, str_var, label, command):
            self.menu_options.add_command(label=label, state='disabled')
            for strategy in strategy_list:
                self.menu_options.add_radiobutton(label=str(strategy()),
                value=strategy.name, variable=str_var, command=command)
            self.menu_options.add_separator()  # separators if necessary

        make_options_submenu(
            strategy_list=ReadingStrategies.registry.values(),
            str_var=self.reading_strategy,
            label='Reading',
            command=self.set_reading_method
        )
        make_options_submenu(
            strategy_list=BackfittingStrategies.registry.values(),
            str_var=self.backfitting_strategy,
            label='Background fitting',
            command=self.set_backfitting_method
        )
        make_options_submenu(
            strategy_list=PeakfittingStrategies.registry.values(),
            str_var=self.peakfitting_strategy,
            label='Spectrum fitting',
            command=self.set_peakfitting_method
        )
        make_options_submenu(
            strategy_list=CorrectingStrategies.registry.values(),
            str_var=self.correcting_strategy,
            label='Temperature correction',
            command=self.set_correcting_method
        )
        make_options_submenu(
            strategy_list=TranslatingStrategies.registry.values(),
            str_var=self.translating_strategy,
            label='Pressure determination',
            command=self.set_translating_method
        )
        make_options_submenu(
            strategy_list=DrawingStrategies.registry.values(),
            str_var=self.drawing_strategy,
            label='Drawing style',
            command=self.set_drawing_method
        )
        self.menu.add_command(label='?', command=show_about)

        # BUILDING ENTRY AREA
        self.file = FilenameEntry(self, self.file_to_previous,
                                  self.file_from_entry, self.file_to_next)
        self.file.grid(row=0, column=0, columnspan=4)
        self.r1 = UfloatEntry(self, label='R1', entry=self.r1_ref,
                              unit='nm', cmd=self.recalculate_p)
        self.r1.grid(row=1, column=0, columnspan=4)
        self.t = UfloatEntry(self, label='T', entry=self.t_ref, offset=273.15,
                             unit='\u00B0C', cmd=self.recalculate_p)
        self.t.grid(row=2, column=0, columnspan=4)
        self.p = UfloatEntry(self, label='p', entry=self.t_ref,
                             unit='GPa', cmd=self.recalculate_r)
        self.p.grid(row=3, column=0, columnspan=4)

        # BUILDING STATUS BAR
        self.status_bar = StatusBar(self)
        self.status_bar.grid(row=4, column=0, columnspan=4)

        # FINAL SETUP
        self.set_all_methods()
        self.save_reference()  # ran to update correlation b/ temp corr and r1!
        self.display('Ready...')

    def display(self, message):
        self.status_bar.display(message)

    def recalculate_p(self, *_):
        # calculate the pressure for current data
        self.calc.r1 = self.r1.get()
        self.calc.t = self.t.get()
        self.calc.calculate_p_from_r1()
        self.p.set(value=self.calc.p)
        if self.autodraw.get() is True:
            self.data_draw()
        self.display('Calculated p from R1 and T.')

    def recalculate_r(self, *_):
        self.calc.p = self.p.get()
        self.calc.t = self.t.get()
        self.calc.calculate_r1_from_p()
        self.r1.set(self.calc.r1)
        self.display('Calculated R1 from p and T.')

    def open_file_dialogue(self):
        path = open_file_dialogue()
        if not path:
            self.display('No file have been loaded.')
            return
        new_directory, new_filename = os.path.split(path)
        os.chdir(new_directory)
        self.change_file(new_filename)

    def change_file(self, filename):
        self.file.set(filename)
        if filename == '':
            return
        try:
            self.calc.read(filename)
        except RuntimeError:
            self.display('Fitting spectrum failed!')
            return
        except FileNotFoundError:
            self.display('File does not exist!')
            return
        except (TypeError, ValueError, UnicodeDecodeError):
            self.display('File cannot be intepreted!')
            return
        except PermissionError:
            self.display('No permissions to read this file!')
            return
        self.r1.set(value=self.calc.r1)
        self.recalculate_p()

    def file_to_next(self):
        return self.change_file(self.get_filename(self.file.get(), shift=+1))

    def file_to_previous(self):
        return self.change_file(self.get_filename(self.file.get(), shift=-1))

    def file_from_entry(self, *_):
        filename = self.file.get()
        if os.path.isfile(os.path.join(os.getcwd(), filename)):
            self.change_file(filename)
        elif os.path.isfile(filename):
            self.change_file(filename)

    def save_reference(self):
        self.calc.r1 = self.r1_ref = self.r1.get()
        self.calc.t = self.t_ref = self.t.get()
        self.calc.p = self.p_ref = self.p.get()
        self.calc.set_current_as_reference()
        self.recalculate_p()
        self.display('R1 & T saved as reference for p=0.')

    def load_reference(self):
        self.r1.set(self.r1_ref)
        self.t.set(self.t_ref)
        self.recalculate_p()
        self.display('R1 & T loaded from reference.')

    @staticmethod
    def get_filename(filename, shift=0):
        files = [f for f in os.listdir(os.getcwd()) if not os.path.isdir(f)]
        files = natsorted(files)
        try:
            index = files.index(filename)
        except ValueError:
            index, shift = 0, 0
        return files[(len(files) + index + shift) % len(files)]

    def set_all_methods(self):
        self.calc.engine.set_strategy(
            reading=self.reading_strategy.get(),
            backfitting=self.backfitting_strategy.get(),
            peakfitting=self.peakfitting_strategy.get(),
            correcting=self.correcting_strategy.get(),
            translating=self.translating_strategy.get(),
            drawing=self.drawing_strategy.get())
        self._reevaluate()

    def set_reading_method(self):
        self.calc.engine.set_strategy(reading=self.reading_strategy.get())
        self._reevaluate()

    def set_backfitting_method(self):
        self.calc.engine.set_strategy(backfitting=self.backfitting_strategy.get())
        self._reevaluate()

    def set_peakfitting_method(self):
        self.calc.engine.set_strategy(peakfitting=self.peakfitting_strategy.get())
        self._reevaluate()

    def set_correcting_method(self):
        self.calc.engine.set_strategy(correcting=self.correcting_strategy.get())
        self._reevaluate()

    def set_translating_method(self):
        self.calc.engine.set_strategy(translating=self.translating_strategy.get())
        self._reevaluate()

    def set_drawing_method(self):
        self.calc.engine.set_strategy(drawing=self.drawing_strategy.get())
        self._reevaluate()

    def _reevaluate(self):
        self.calc.calculate_offset_from_reference()
        try:
            self.calc.read()
            self.r1.set(value=self.calc.r1)
        except OSError:
            pass
        self.recalculate_p()

    def data_draw(self):
        try:
            self.calc.draw()
            self.display('Drawn currently loaded data.')
        except AttributeError:
            self.display('No loaded spectrum to draw!')



import unittest
from math import pi, inf
from pruby.utility import cycle, LineSubset
from pruby.utility import polynomial, gaussian, lorentzian, pseudovoigt


class TestCycle(unittest.TestCase):
    def test_creation(self):
        self.assertTrue(cycle('contents'))

    def test_single_yield(self):
        self.assertTrue(next(cycle('contents')))

    def test_multi_yield(self):
        for index in range(100):
            self.assertTrue(next(cycle('contents')))

    def test_yield_order(self):
        c = cycle(list('contents'))
        self.assertEqual(next(c) + next(c) + next(c), 'con')

    def test_start(self):
        c = cycle(list('contents'), start=3)
        self.assertEqual(next(c) + next(c) + next(c), 'ten')


class TestFunctions(unittest.TestCase):
    def test_empty_polynomial(self):
        self.assertAlmostEqual(polynomial()(pi), 0.0)

    def test_empty_pseudovoigt(self):
        insufficient_arguments = 12., -34., 56.
        self.assertRaises(TypeError, pseudovoigt, insufficient_arguments)

    def test_polynomial_callable(self):
        self.assertTrue(callable(polynomial(12., -34., 56.)))

    def test_gaussian_callable(self):
        self.assertTrue(callable(gaussian(12., -34., 56.)))

    def test_lorenzian_callable(self):
        self.assertTrue(callable(lorentzian(12., -34., 56.)))

    def test_pseudovoigt_callable(self):
        self.assertTrue(callable(pseudovoigt(12., -34., 56., 78.)))

    def test_polynomial_value(self):
        self.assertAlmostEqual(polynomial(12., -34., 56.)(pi), 457.8836962389)

    def test_gaussian_value(self):
        self.assertAlmostEqual(gaussian(12., -34., 56.)(pi), 9.6307508535)

    def test_lorenzian_value(self):
        self.assertAlmostEqual(lorentzian(12., -34., 56.)(pi), 8.3339646686)

    def test_pseudovoigt_value(self):
        self.assertAlmostEqual(pseudovoigt(12, -34, 56, 78)(pi), -58.3996662756)


class TestLineSubset(unittest.TestCase):
    def test_create_from_pair(self):
        self.assertTrue(LineSubset(1.2, inf))

    def test_create_from_list(self):
        self.assertTrue(LineSubset([(-inf, -2), (-1, 1), (1.2, inf)]))

    def test_create_empty(self):
        self.assertTrue(LineSubset())

    def test_create_incorrect(self):
        with self.assertRaises(ValueError):
            LineSubset(3, 2)

    def test_merging(self):
        self.assertEqual(LineSubset([(0, 1), (1, 2), (2, 3)]), LineSubset(0, 3))

    def test_negation(self):
        self.assertEqual(LineSubset([(-inf, -2.2), (3.3, inf)]),
                         -LineSubset(-2.2, 3.3))

    def test_empty_infinity_negation(self):
        self.assertEqual(-(-LineSubset()), LineSubset())

    def test_equal(self):
        self.assertTrue(LineSubset(1.1, 2.2) == LineSubset([(1.1, 2.2)]))

    def test_lesser_than(self):
        self.assertTrue(LineSubset(2.2, 3.3) < LineSubset([(1.1, 3.3), (4, 5)]))

    def test_lesser_equal(self):
        self.assertTrue(LineSubset(1.1, 3.3) <= LineSubset(1.1, 3.3))

    def test_union(self):
        self.assertEqual(LineSubset(1.1, 3.3) + LineSubset(2.2, inf),
                         LineSubset(1.1, inf))

    def test_intersection(self):
        self.assertEqual(LineSubset(-inf, 2.2) * LineSubset(1.1, 3.3),
                         LineSubset(1.1, 2.2))

    def test_subtraction(self):
        self.assertEqual(LineSubset(-inf, 2.2) - LineSubset(1.1, 3.3),
                         LineSubset(-inf, 1.1))

    def test_division(self):
        self.assertEqual(LineSubset(-inf, 2.2) / LineSubset(1.1, 3.3),
                         LineSubset([(-inf, 1.1), (2.2, 3.3)]))

    def test_representation(self):
        self.assertIn('2.2, inf', repr(LineSubset(2.2, inf)))

    def test_string(self):
        self.assertIn('2.2, inf', str(LineSubset(2.2, inf)))

    def test_max(self):
        self.assertEqual(max(LineSubset(-inf, -2.2)), -2.2)

    def test_iterator(self):
        self.assertEqual(list(LineSubset([(0, 1), (2, 3)])), [0, 1, 2, 3])

    def test_contains_float(self):
        self.assertIn(1.1, LineSubset(1.1, inf))

    def test_contains_other(self):
        self.assertIn(LineSubset(1.1, 3.3), LineSubset(1.1, inf))


if __name__ == '__main__':
    unittest.main()



import abc
from collections import UserDict, OrderedDict
from typing import Callable


class BaseStrategy(abc.ABC):
    """Base class for all types of all individual strategies"""

    def __str__(self):
        return f'{self.name} ({self.year})' if self.year else self.name

    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    def year(self) -> int:
        return 0

    @property
    def reference(self) -> str:
        return ''


class BaseStrategies(UserDict, abc.ABC):
    """Abstract class holding individual strategies as a name: strategy dict."""

    strategy_type = BaseStrategy
    default: strategy_type = None
    registry: OrderedDict

    @abc.abstractmethod
    def registry(self) -> dict:
        raise NotImplementedError

    @classmethod
    def create(cls, name: str = '') -> strategy_type:
        strategy_class = cls.registry[name] if name else cls.default
        return strategy_class()

    @classmethod
    def register(cls, default: bool = False) -> Callable:
        def decorator(decorated_strategy: cls.strategy_type)\
                -> cls.strategy_type:
            name = decorated_strategy.name
            if name in cls.registry:
                raise KeyError(f'{name} already registered in {cls.__name__}')
            if default:
                if cls.default is not None:
                    raise ValueError(f'2nd default {name} for {cls.__name__}')
                cls.default = decorated_strategy
            cls.registry[name] = decorated_strategy
            return decorated_strategy
        return decorator



import pathlib
import tempfile
import unittest
from pruby.engine import Engine
from pruby import PressureCalculator
from pruby import strategies


test_data1_path = str(pathlib.Path(__file__).parent.joinpath('test_data1.txt'))
test_data2_path = str(pathlib.Path(__file__).parent.joinpath('test_data2.txt'))

subengines = \
    [
        'reader',
        'backfitter',
        'peakfitter',
        'corrector',
        'translator',
        'drawer'
    ]
strategy_types = \
    [
        'reading',
        'backfitting',
        'peakfitting',
        'correcting',
        'translating',
        'drawing'
    ]
strategy_families = \
    [
        strategies.ReadingStrategies,
        strategies.BackfittingStrategies,
        strategies.PeakfittingStrategies,
        strategies.CorrectingStrategies,
        strategies.TranslatingStrategies,
        strategies.DrawingStrategies
    ]
strategy_parents = \
    [
        strategies.ReadingStrategy,
        strategies.BackfittingStrategy,
        strategies.PeakfittingStrategy,
        strategies.CorrectingStrategy,
        strategies.TranslatingStrategy,
        strategies.DrawingStrategy
    ]


class TestEngine(unittest.TestCase):
    def test_create_empty(self):
        self.assertIsNotNone(Engine(calc=PressureCalculator()))

    def test_has_default_strategies(self):
        e = Engine(calc=PressureCalculator())
        for subengine in subengines:
            self.assertTrue(hasattr(e, subengine))

    def test_can_set_null_strategies(self):
        s = Engine(calc=PressureCalculator())
        s.set_strategy()
        for sub in subengines:
            self.assertIsInstance(getattr(s, sub), strategies.BaseStrategy)

    def test_can_set_existing_strategies(self):
        s = Engine(calc=PressureCalculator())
        for sub, st, sf in zip(subengines, strategy_types, strategy_families):
            for strategy_name in sf.registry.keys():
                s.set_strategy(**{st: strategy_name})
                self.assertIsInstance(getattr(s, sub), strategies.BaseStrategy)

    def test_cant_set_unexisting_strategies(self):
        engine = Engine(calc=PressureCalculator())
        for sub in subengines:
            with self.assertRaises(TypeError):
                engine.set_strategy(**{sub: f'dummy {sub} name'})

    def test_can_assign_existing_subengines(self):
        s = Engine(calc=PressureCalculator())
        for sub, st, sf in zip(subengines, strategy_types, strategy_families):
            for strategy in sf.registry.values():
                setattr(s, sub, strategy())
                self.assertIsInstance(getattr(s, sub), strategies.BaseStrategy)


# noinspection PyTypeChecker
class TestCalculator(unittest.TestCase):
    def test_create_empty(self):
        self.assertIsNotNone(PressureCalculator())

    def test_zero_pressure_on_start(self):
        calc = PressureCalculator()
        calc.calculate_p_from_r1()
        self.assertAlmostEqual(calc.p.n, 0.0, places=1)

    def test_pressure_changes_with_input(self):
        calc = PressureCalculator()
        p1 = calc.p
        calc.r1 = calc.r1 + 0.1
        calc.calculate_p_from_r1()
        p2 = calc.p
        calc.t = calc.t + 1.0
        calc.calculate_p_from_r1()
        p3 = calc.p
        self.assertNotAlmostEqual(p1.n, p2.n)
        self.assertNotAlmostEqual(p2.n, p3.n)

    def test_reference_changes(self):
        calc = PressureCalculator()
        calc.r1 = calc.r1 + 0.1
        calc.t = calc.t + 1.0
        calc.set_current_as_reference()
        calc.calculate_p_from_r1()
        self.assertAlmostEqual(calc.p.n, 0.0, places=4)

    def test_r1_calculations_in_modest_range(self):
        calc = PressureCalculator()
        calc.p = calc.p + 100.0
        calc.calculate_r1_from_p()
        self.assertGreater(calc.r1, 710.0)
        calc.p = calc.p - 110.0
        calc.calculate_r1_from_p()
        self.assertLess(calc.r1, 693.0)

    def test_reading_and_fitting_simple_raw_file(self):
        calc = PressureCalculator()
        calc.read(test_data1_path)
        self.assertGreater(calc.r1.n, 694.0)

    def test_reading_and_fitting_meta_file(self):
        calc = PressureCalculator()
        calc.engine.set_strategy(reading='Metadata spectrum txt')
        calc.read(test_data2_path)
        self.assertGreater(calc.r1.n, 694.0)

    def test_different_fitters_give_different_r1(self):
        calc1 = PressureCalculator()
        calc2 = PressureCalculator()
        calc3 = PressureCalculator()
        calc1.engine.set_strategy(backfitting='Linear Huber',
                                  peakfitting='Gaussian')
        calc2.engine.set_strategy(backfitting='Linear Satelite',
                                  peakfitting='Gaussian')
        calc3.engine.set_strategy(backfitting='Linear Huber',
                                  peakfitting='Camel')
        calc1.read(test_data1_path)
        calc2.read(test_data1_path)
        calc3.read(test_data1_path)
        self.assertNotAlmostEqual(calc1.r1.n, calc2.r1.n)
        self.assertNotAlmostEqual(calc1.r1.n, calc3.r1.n)
        self.assertNotAlmostEqual(calc2.r1.n, calc3.r1.n)

    def test_different_correctors_translators_give_different_p(self):
        calc1 = PressureCalculator()
        calc2 = PressureCalculator()
        calc3 = PressureCalculator()
        calc4 = PressureCalculator()
        calc1.r1, calc1.t = calc1.r1 + 2.0, calc1.t + 2.0
        calc2.r1, calc2.t = calc2.r1 + 2.0, calc2.t + 2.0
        calc3.r1, calc3.t = calc3.r1 + 2.0, calc3.t + 2.0
        calc4.r1, calc4.t = calc4.r1 + 2.0, calc4.t + 2.0
        calc1.engine.set_strategy(correcting='Vos R1', translating='Liu')
        calc2.engine.set_strategy(correcting='Ragan R1', translating='Mao')
        calc3.engine.set_strategy(correcting='None', translating='Piermarini')
        calc4.engine.set_strategy(correcting='None', translating='Wei')
        calc1.calculate_p_from_r1()
        calc2.calculate_p_from_r1()
        calc3.calculate_p_from_r1()
        calc4.calculate_p_from_r1()
        self.assertNotAlmostEqual(calc1.p.n, calc2.p.n)
        self.assertNotAlmostEqual(calc1.p.n, calc3.p.n)
        self.assertNotAlmostEqual(calc1.p.n, calc4.p.n)
        self.assertNotAlmostEqual(calc2.p.n, calc3.p.n)
        self.assertNotAlmostEqual(calc2.p.n, calc4.p.n)
        self.assertNotAlmostEqual(calc3.p.n, calc4.p.n)

    def test_drawing(self):
        calc = PressureCalculator()
        calc.read(test_data2_path)
        for strategy in strategies.DrawingStrategies.registry.keys():
            with tempfile.TemporaryDirectory() as temp_dir:
                png_path = temp_dir + 'file.png'
                calc.output_path = png_path
                calc.engine.set_strategy(drawing=strategy)
                calc.draw()
                self.assertTrue(pathlib.Path(png_path).is_file())


if __name__ == '__main__':
    unittest.main()



import matplotlib.pyplot as plt
import uncertainties as uc
from pruby.engine import Engine
from pruby.spectrum import Spectrum
from pruby.utility import LineSubset
from pruby.constants import P_0, R1_0, R2_0, T_0, UZERO


class PressureCalculator:
    def __init__(self):
        self.engine: Engine = Engine(calc=self)
        self.dat_path: str = ''
        self.ref_path: str = ''
        self.limits: LineSubset = LineSubset(690.0, 705.0)
        self.raw_spectrum: Spectrum = Spectrum()
        self.back_spectrum: Spectrum = Spectrum()
        self.peak_spectrum: Spectrum = Spectrum()
        self.r1: uc.UFloat = R1_0
        self.r2: uc.UFloat = R2_0
        self.r1_ref: uc.UFloat = R1_0
        self.t: uc.UFloat = T_0
        self.t_ref: uc.UFloat = T_0
        self.p: uc.UFloat = P_0
        self.t_correction: uc.UFloat = UZERO
        self.offset: uc.UFloat = UZERO
        self.fig: plt.Figure = plt.Figure()
        self.output_path: str = ''
        self.calculate_p_from_r1()

    def set_current_as_reference(self):
        self.r1_ref = self.r1
        self.t_ref = self.t
        self.ref_path = self.dat_path
        self.calculate_p_from_r1()
        self.calculate_offset_from_reference()
        self.calculate_p_from_r1()

    def calculate_offset_from_reference(self):
        backupped_values = self.r1, self.t, self.p
        self.offset = uc.ufloat(0.00, 0.01)
        self.t, self.p = self.t_ref, P_0
        self.calculate_r1_from_p()
        self.offset = self.r1_ref - self.r1
        self.r1, self.t, self.p = backupped_values

    def read(self, path: str = ''):
        self.dat_path = path if path else self.dat_path
        self.engine.read()
        self.engine.backfit()
        self.engine.peakfit()

    def calculate_p_from_r1(self):
        self.engine.correct()
        self.engine.translate()

    def calculate_r1_from_p(self):
        target_p = self.p
        self.p = self.p - uc.ufloat(100, 0)
        precision = 0.0001
        r_step_size = uc.ufloat(1.0, 0)
        self.r1 = uc.ufloat(self.r1.n, precision)
        for _ in range(100):
            previous_p = self.p
            self.calculate_p_from_r1()
            if abs(target_p - self.p) < precision:
                break
            if abs(self.p - previous_p) > abs(previous_p - target_p):
                r_step_size = r_step_size * 0.5
            if self.p < target_p:
                self.r1 += r_step_size
            else:
                self.r1 -= r_step_size

    def draw(self):
        if self.peak_spectrum:
            self.engine.draw()
        else:
            raise AttributeError('The peak spectrum is empty')



from pruby.strategies import \
    ReadingStrategies, \
    BackfittingStrategies, \
    PeakfittingStrategies, \
    CorrectingStrategies, \
    TranslatingStrategies, \
    DrawingStrategies


class Engine:
    def __init__(self, calc):
        self.calc = calc
        self.reader = ReadingStrategies.default()
        self.backfitter = BackfittingStrategies.default()
        self.peakfitter = PeakfittingStrategies.default()
        self.corrector = CorrectingStrategies.default()
        self.translator = TranslatingStrategies.default()
        self.drawer = DrawingStrategies.default()

    def set_strategy(self, reading: str = '', backfitting: str = '',
                     peakfitting: str = '', correcting: str = '',
                     translating: str = '', drawing: str = '') -> None:
        """
        Sets engine strategy using strategy name string. To change any of the
        strategies directly using a `Strategy` object, set the value of one of:
        `self.reader`, `self.backfitter`, `self.peakfitter`, `self.corrector`,
        `self.translator` or `self.drawer` to a desired class instance instead.

        :param reading: If given, set `self.reader` to an instance
            of class registered in `ReadingStrategies` under this name.
        :param backfitting: If given, set `self.backfitter` to an instance
            of class registered in `BackfittingStrategies` under this name.
        :param peakfitting: If given, set `self.peakfitter` to an instance
            of class registered in `PeakfittingStrategies` under this name.
        :param correcting: If given, set `self.corrector` to an instance
            of class registered in `CorrectingStrategies` under this name.
        :param translating: If given, set `self.translator` to an instance
            of class registered in `TranslatingStrategies` under this name.
        :param drawing: If given, set `self.drawer` to an instance
            of class registered in `DrawingStrategies` under this name.
        """
        if reading:
            self.reader = ReadingStrategies.create(name=reading)
        if backfitting:
            self.backfitter = BackfittingStrategies.create(name=backfitting)
        if peakfitting:
            self.peakfitter = PeakfittingStrategies.create(name=peakfitting)
        if correcting:
            self.corrector = CorrectingStrategies.create(name=correcting)
        if translating:
            self.translator = TranslatingStrategies.create(name=translating)
        if drawing:
            self.drawer = DrawingStrategies.create(name=drawing)

    def read(self):
        self.reader.read(self.calc)

    def backfit(self):
        self.backfitter.backfit(self.calc)

    def peakfit(self):
        self.peakfitter.peakfit(self.calc)

    def correct(self):
        self.corrector.correct(self.calc)

    def translate(self):
        self.translator.translate(self.calc)

    def draw(self):
        self.drawer.draw(self.calc)



def xmin(obj):
    try:
        return min(obj)
    except TypeError:
        return obj


def xmax(obj):
    try:
        return max(obj)
    except TypeError:
        return obj


class LineSubset:

    # SEGMENT METHODS
    class LineSegment:
        def __init__(self, left=None, right=None):
            if left > right:
                raise ValueError('Left ({}) > Right ({})'.format(left, right))
            self.left = left
            self.right = right

        def __eq__(self, other):
            return self.left == other.left and self.right == other.right

        def __ne__(self, other):
            return not(self == other)

        def __str__(self):
            return '[{}, {}]'.format(self.left, self.right)

        def __iter__(self):
            yield self.left
            yield self.right

        def __contains__(self, item):
            return all((xmin(self) <= xmin(item), xmax(self) >= xmax(item)))

    # CREATION METHODS
    def __init__(self, left=None, right=None):
        self.segments = []
        if left is None and right is None:
            pass
        elif right is None:
            for left1, left2 in left:
                self.segments.append(self.LineSegment(left1, left2))
        else:
            self.segments.append(self.LineSegment(left, right))
        self.segments = self._merge(self.segments)

    def _merge(self, segments):
        limits = sorted(list(zip(list(self), ['l', 'r'] * len(self.segments))))
        segments = []
        while limits:
            left = limits.pop(0)[0]
            depth = 1
            while limits:
                lim = limits.pop(0)
                if lim[1] == 'l':
                    depth += 1
                elif lim[1] == 'r' and depth > 1:
                    depth -= 1
                else:
                    segments.append(self.LineSegment(left, lim[0]))
                    break
        return segments

    # COMPARISON METHODS
    def __eq__(self, other):
        return self.segments == other.segments

    def __ne__(self, other):
        return not(self == other)

    def __lt__(self, other):
        return all(any(s in o for o in other.segments) for s in self.segments) \
               and not(self == other)

    def __gt__(self, other):
        return other < self

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return other < self or self == other

    # UNARY OPERATIONS
    def __pos__(self):
        return self

    def __neg__(self):
        limits = [-float('inf')] + list(self) + [float('inf')]
        lefts = limits[0::2]
        rights = limits[1::2]
        segments = [(lef, rig) for lef, rig in zip(lefts, rights) if lef < rig]
        return self.__class__(segments)

    # ARITHMETIC METHODS
    def __add__(self, other):
        segments = self.segments + other.segments
        return self.__class__(segments)

    def __sub__(self, other):
        return self * -other

    def __mul__(self, other):
        return -(-self + -other)

    def __truediv__(self, other):
        return (self - other) + (other - self)

    # REPRESENTATION METHODS
    def __str__(self):
        return ' + '.join([str(s) for s in self.segments])

    def __repr__(self):
        zipped = zip(list(self)[0::2], list(self)[1::2])
        pairs = ['{}, {}'.format(l, r) for l, r in zipped]
        return 'LineSubset({})'.format('; '.join(pairs))

    # CONTAINER METHODS
    def __iter__(self):
        return iter([lim for pair in self.segments for lim in pair])

    def __contains__(self, item):
        return any(item in segment for segment in self.segments)



import abc
import copy
from collections import OrderedDict
from scipy.optimize import curve_fit
from pruby.strategies import BaseStrategy, BaseStrategies
from pruby.utility import polynomial
from pruby.spectrum import Curve


class BackfittingStrategy(BaseStrategy, abc.ABC):
    @abc.abstractmethod
    def backfit(self, calc):
        raise NotImplementedError


class BackfittingStrategies(BaseStrategies):
    registry = OrderedDict()
    strategy_type = BackfittingStrategy


class BaseBackfittingStrategy(BackfittingStrategy, abc.ABC):
    reference = r'https://doi.org/10.1016/j.chemolab.2004.10.003'

    @staticmethod
    def _approximate_linearly(spectrum):
        def linear_function(x, _a0, _a1):
            return polynomial(_a0, _a1)(x)
        a1 = (spectrum.y[-1] - spectrum.y[0]) / (spectrum.x[-1] - spectrum.x[0])
        a0 = spectrum.y[0] - a1 * spectrum.x[0]
        spectrum.curve = Curve(func=linear_function, args=(a0, a1))

    @abc.abstractmethod
    def _prepare_backfit(self, calc):
        pass

    def backfit(self, calc):
        self._prepare_backfit(calc)
        for cycle in range(50):
            previous_mse = calc.back_spectrum.mse
            x = calc.back_spectrum.focused.x
            y = calc.back_spectrum.focused.y
            si = calc.back_spectrum.focused.si
            calc.back_spectrum.curve.args, _ = \
                curve_fit(calc.back_spectrum.curve, xdata=x, ydata=y,
                          p0=calc.back_spectrum.curve.args, sigma=si)
            if previous_mse / calc.back_spectrum.mse - 1 < 1e-10:
                break
        calc.back_spectrum.y = calc.back_spectrum.f
        calc.peak_spectrum = copy.deepcopy(calc.raw_spectrum)
        calc.peak_spectrum.y = calc.raw_spectrum.y - calc.back_spectrum.y


@BackfittingStrategies.register(default=True)
class HuberBackfittingStrategy(BaseBackfittingStrategy):
    name = 'Linear Huber'

    def _prepare_backfit(self, calc):
        calc.back_spectrum = copy.deepcopy(calc.raw_spectrum)
        self._approximate_linearly(calc.back_spectrum)
        calc.back_spectrum.focus_on_whole()
        calc.back_spectrum.sigma_type = 'huber'


@BackfittingStrategies.register()
class SateliteBackfittingStrategy(BaseBackfittingStrategy):
    name = 'Linear Satelite'

    def _prepare_backfit(self, calc):
        calc.back_spectrum = copy.deepcopy(calc.raw_spectrum)
        self._approximate_linearly(calc.back_spectrum)
        calc.back_spectrum.focus_on_edge(width=1.0)
        calc.back_spectrum.sigma_type = 'equal'


@BackfittingStrategies.register()
class NullBackfittingStrategy(BackfittingStrategy):
    name = 'No background fitting'

    def backfit(self, calc):
        calc.back_spectrum = copy.deepcopy(calc.raw_spectrum)
        calc.back_spectrum.y = calc.back_spectrum.y * 0.0



import tkinter as tk
from pruby.gui.app import Application
from pruby.resources import icon


def run():
    root = tk.Tk()
    Application(root).pack(side='top', fill='both', expand=True)
    root.title('pRuby')
    root.attributes("-topmost", True)
    root.tk.call('wm', 'iconphoto', root._w, tk.PhotoImage(data=icon))
    root.resizable(False, False)
    root.mainloop()



import enum
import numpy as np

from .curve import Curve
from ..utility.line_subset import LineSubset


class Spectrum:
    class SigmaType(enum.Enum):
        equal = 'equal'
        huber = 'huber'

    def __init__(self, x=tuple(), y=tuple(), curve=Curve(),
                 focus=LineSubset(), sigma_type='equal'):
        self.x = np.array(x)
        self.y = np.array(y)
        self.curve = curve
        self.focus = self.domain if focus == LineSubset() else focus
        self.sigma_type = sigma_type

    def __len__(self):
        return len(self.x)

    def __bool__(self):
        return len(self.x) > 0

    @property
    def sigma_type(self):
        return self._sigma_type

    @sigma_type.setter
    def sigma_type(self, value):
        self._sigma_type = self.SigmaType(value)

    @property
    def f(self):
        return np.array(list(map(self.curve, self.x)))

    @property
    def delta(self):
        return self.y - self.f

    @property
    def si(self):
        if self.sigma_type == self.SigmaType.equal:
            return np.ones_like(self.x)
        elif self.sigma_type == self.SigmaType.huber:
            tol = 0.01 * max(abs(self.delta))
            return np.array([tol ** 2 if d < tol else
                             tol * (2 * d - tol) for d in self.delta])
        else:
            raise KeyError('Unknown sigma type "{}"'.format(self.sigma_type))

    @property
    def mse(self):
        return sum((self.delta / self.si) ** 2) / len(self.delta)

    @property
    def domain(self):
        if len(self.x) == 0:
            return LineSubset()
        else:
            return LineSubset(min(self.x), max(self.x))

    @property
    def focused(self):
        return self.within(self.focus)

    def within(self, subset):
        in_subset = [True if x in subset else False for x in self.x]
        return Spectrum(self.x[in_subset], self.y[in_subset], curve=self.curve,
                        focus=subset, sigma_type=self.sigma_type.value)

    def focus_on_edge(self, width=1.0):
        sub1 = LineSubset(min(self.x), min(self.x) + width)
        sub2 = LineSubset(max(self.x) - width, max(self.x))
        self.focus = sub1 + sub2

    def focus_on_whole(self):
        self.focus = self.domain

    def focus_on_points(self, points, width=1.0):
        points_area = LineSubset([(p - width/2, p + width/2) for p in points])
        self.focus = points_area * self.domain



import abc
import numpy as np
from collections import OrderedDict
from uncertainties import ufloat_fromstr
from pruby.strategies.base import BaseStrategy, BaseStrategies
from pruby.spectrum import Spectrum


class ReadingStrategy(BaseStrategy, abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def read(calc):
        raise NotImplementedError


class ReadingStrategies(BaseStrategies):
    registry = OrderedDict()
    strategy_type = ReadingStrategy


@ReadingStrategies.register()
class RawTxtReadingStrategy(ReadingStrategy):
    name = 'Raw spectrum txt'

    @staticmethod
    def read(calc):
        data = np.loadtxt(calc.dat_path, dtype=(float, float))
        x_list, y_list = data[:, 0], data[:, 1]
        calc.raw_spectrum = Spectrum(x_list, y_list).within(calc.limits)


@ReadingStrategies.register(default=True)
class MetaTxtReadingStrategy(ReadingStrategy):
    name = 'Metadata spectrum txt'

    @staticmethod
    def read(calc):
        with open(calc.dat_path, 'r') as file:
            x_list, y_list = [], []
            for line in file.readlines():
                try:
                    split_line = line.strip().split()
                    new_x = float(split_line[0])
                    new_y = float(split_line[1])
                except ValueError:
                    pass
                else:
                    x_list.append(new_x)
                    y_list.append(new_y)
        calc.raw_spectrum = Spectrum(x_list, y_list).within(calc.limits)


@ReadingStrategies.register()
class SingleLineReadingStrategy(ReadingStrategy):
    name = 'Single value txt'

    @staticmethod
    def read(calc):
        with open(calc.dat_path, 'r') as file:
            first_line = file.readline().split()
            r1 = ufloat_fromstr(first_line[0])
            try:
                intensity = ufloat_fromstr(first_line[1])
            except (IndexError, ValueError):
                intensity = 1
        calc.raw_spectrum = calc.peak_spectrum = Spectrum([r1], [intensity])
        calc.r1 = r1



import tkinter as tk
import uncertainties as uc


PADX = 1
PADY = 0
WIDTH = 24


class FilenameEntry(tk.Frame):
    def __init__(self, parent, left_cmd, entry_cmd, right_cmd):
        tk.Frame.__init__(self, parent)
        self.value = tk.StringVar(value='')
        prev_button = tk.Button(self, text='\u25c0', command=left_cmd)
        prev_button.grid(row=0, column=0, padx=PADX, pady=PADY)
        self.entry = tk.Entry(self, width=int(0.7 * WIDTH), justify='right')
        self.set('')
        self.entry.grid(row=0, column=1, columnspan=2, padx=PADX, pady=PADY)
        self.entry.bind('<Return>', entry_cmd)
        next_button = tk.Button(self, text='\u25b6', command=right_cmd)
        next_button.grid(row=0, column=3, padx=PADX, pady=PADY)

    def get(self):
        return self.entry.get()

    def set(self, value):
        self.entry.delete(0, tk.END)
        self.entry.insert(0, value)


class UfloatEntry(tk.Frame):
    def __init__(self, parent, label, entry, unit, cmd, offset=0.0):
        tk.Frame.__init__(self, parent)
        self.parent = parent
        self.offset = offset  # to convert celsius into kelvin
        self.label = tk.Label(self, text=label)
        self.label.config(width=int(0.15*WIDTH), justify=tk.CENTER)
        self.label.grid(row=1, column=0, padx=PADX, pady=PADY)
        self.entry = tk.Entry(self, width=int(0.7*WIDTH), justify='right')
        self.set(entry)
        self.entry.grid(row=1, column=1, columnspan=2, padx=PADX, pady=PADY)
        self.entry.bind('<Return>', cmd)
        self.unit = tk.Label(self, text=unit)
        self.unit.config(width=int(0.15*WIDTH), justify=tk.CENTER)
        self.unit.grid(row=1, column=3, padx=PADX, pady=PADY)

    def get(self):
        try:
            uc.ufloat_fromstr(self.entry.get())
            return uc.ufloat_fromstr(self.entry.get()) + self.offset
        except ValueError as error:
            self.parent.display('Incorrect entry input format!')
            raise ValueError(error)

    def set(self, value):
        self.entry.delete(0, tk.END)
        self.entry.insert(0, '{0:.2uS}'.format(value - self.offset))


class StatusBar(tk.Frame):
    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)
        self.bar = tk.Label(self, bd=1, relief=tk.SUNKEN, width=WIDTH+2*PADX)
        self.bar.config(anchor='w')
        self.bar.pack()

    def display(self, message):
        self.bar.config(text=message)

    def grid(self, **kwargs):
        kwargs['padx'] = 0
        kwargs['pady'] = 1
        kwargs['sticky'] = 'NSWE'
        super().grid(**kwargs)



import abc
from collections import OrderedDict
from pruby.strategies.base import BaseStrategy, BaseStrategies
from uncertainties import ufloat
from pruby.constants import R1_0


def mao_function(r1, a, b):
    """Based on doi:10.1063/1.325277"""
    return (a / b) * (((r1 / R1_0) ** b) - 1)


class TranslatingStrategy(BaseStrategy, abc.ABC):
    @abc.abstractmethod
    def translate(self, calc):
        raise NotImplementedError


class TranslatingStrategies(BaseStrategies):
    registry = OrderedDict()
    strategy_type = TranslatingStrategy


@TranslatingStrategies.register()
class JacobsenTranslatingStrategy(TranslatingStrategy):
    name = 'Jacobsen'
    year = 2008
    reference = r'https://doi.org/10.2138/am.2008.2988'

    def translate(self, calc):
        r1 = calc.r1 - calc.offset + calc.t_correction
        calc.p = mao_function(r1, a=1904, b=ufloat(10.32, 0.07))


@TranslatingStrategies.register()
class LiuTranslatingStrategy(TranslatingStrategy):
    """Based on doi:10.1088/1674-1056/22/5/056201"""
    name = 'Liu'
    year = 2013
    reference = r'https://doi.org/10.1088/1674-1056/22/5/056201'

    def translate(self, calc):
        r1 = calc.r1 - calc.offset + calc.t_correction
        calc.p = mao_function(r1, a=1904, b=9.827)


@TranslatingStrategies.register()
class MaoTranslatingStrategy(TranslatingStrategy):
    name = 'Mao'
    year = 1986
    reference = r'https://doi.org/10.1029/JB091iB05p04673'

    def translate(self, calc):
        r1 = calc.r1 - calc.offset + calc.t_correction
        calc.p = mao_function(r1, a=1904, b=7.665)


@TranslatingStrategies.register()
class PiermariniTranslatingStrategy(TranslatingStrategy):
    name = 'Piermarini'
    year = 1975
    reference = r'https://doi.org/10.1063/1.321957'

    def translate(self, calc):
        r1 = calc.r1 - calc.offset + calc.t_correction
        calc.p = ufloat(2.740, 0.016) * (r1 - R1_0)


@TranslatingStrategies.register(default=True)
class Ruby2020TranslatingStrategy(TranslatingStrategy):
    name = 'Ruby2020'
    year = 2020
    reference = r'https://doi.org/10.1080/08957959.2020.1791107'

    def translate(self, calc):
        r1 = calc.r1 - calc.offset + calc.t_correction
        r1_0_ruby2020 = ufloat(694.25, 0.01)
        r_rel = (r1 - r1_0_ruby2020) / r1_0_ruby2020
        calc.p = ufloat(1870, 10) * r_rel * (1 + ufloat(5.63, 0.03) * r_rel)


@TranslatingStrategies.register()
class WeiTranslatingStrategy(TranslatingStrategy):
    name = 'Wei'
    year = 2011
    reference = r'https://doi.org/10.1063/1.3624618'

    def translate(self, calc):
        r1 = calc.r1 - calc.offset
        a300 = ufloat(1915.0, 0.9)
        a1 = ufloat(0.622, 0.007)
        b300 = ufloat(9.28, 0.02)
        b1 = ufloat(-0.024, 0.003)
        b2 = ufloat(-8.2e-7, 0.02e-7)
        la300 = ufloat(694.2, 0.0)
        la1 = ufloat(0.0063, 0.0002)
        dt = calc.t - 298.0
        a = a300 + a1 * dt
        b = b300 + b1 * dt + b2 * dt ** 2
        la_t = la300 + la1 * dt
        calc.p = (a / b) * (pow(r1 / la_t, b) - 1.0)

# TODO: use years when generating strategy names
# TODO: add doi links or numbers as variables



import unittest
from pruby.resources import icon


class TestIcon(unittest.TestCase):
    def test_icon_exists(self):
        self.assertIsInstance(icon, bytes)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import os
import sys
from setuptools import setup, find_packages

# Version control
python_requires = '>=3.6'
MIN_VERSION = (3, 6)
error_msg = ('This package requires Python %d.%d or higher.' % MIN_VERSION)
try:
    if sys.version_info < MIN_VERSION:
        sys.exit(error_msg)
except AttributeError:  # sys.version_info was introduced in Python 2.0
    sys.exit(error_msg)


# Defining constants, file paths, README contents etc.
HERE = os.path.abspath(os.path.dirname(__file__))
SHORT_DESCRIPTION = 'Python library for calculating pressure ' \
                    'based on ruby fluorescence'
with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = "\n" + f.read()

setup(
    name='pruby',
    version='0.1.3',
    author='Daniel Tchoń',
    author_email='dtchon@chem.uw.edu.pl',
    packages=find_packages(exclude=('legacy', 'cli')),
    package_data={'': ['icon.gif', 'test_data1.txt', 'test_data2.txt']},
    url='https://github.com/Baharis/pRuby',
    license='MIT',
    description=SHORT_DESCRIPTION,
    long_description_content_type='text/markdown',
    long_description=LONG_DESCRIPTION,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    install_requires=[
        'matplotlib>=3.0.0,!=3.3.*',
        'numpy>=1.18.1',
        'scipy>=1.5.1',
        'uncertainties>=3.*',
        'natsort==8.*'
    ]
)



import unittest
from math import pi, sin
from pruby.spectrum import Curve, Spectrum
from pruby.utility import LineSubset


class TestCurve(unittest.TestCase):
    def test_creation_without_arguments(self):
        self.assertTrue(Curve(lambda x: x ** 2))

    def test_creation_with_arguments(self):
        self.assertTrue(Curve(func=lambda x, a, n: a * x ** n, args=(1.2, 3.4)))

    def test_call_without_arguments(self):
        self.assertAlmostEqual(Curve(lambda x: x ** 2)(pi), 9.8696044011)

    def test_call_with_default_arguments(self):
        self.assertAlmostEqual(Curve(lambda x, a: a * x, (1.2,))(pi),
                               3.7699111843)

    def test_call_with_new_arguments(self):
        self.assertAlmostEqual(Curve(lambda x, a: a * x, (1.2,))(pi, 3.4),
                               10.6814150222)

    def test_call_with_no_arguments(self):
        curve = Curve(lambda x, a: a * x, (1.2,))
        with self.assertRaises(IndexError):
            curve()

    def test_call_with_too_many_arguments(self):
        curve = Curve(lambda x, a: a * x, (1.2,))
        with self.assertRaises(IndexError):
            curve(pi, pi, pi)


class TestSpectrum(unittest.TestCase):
    x = [1.0, 2.0, 3.0]
    y = [1.2, 3.4, 5.6]

    def test_creation_of_empty(self):
        self.assertIsNotNone(Spectrum())

    def test_creation_with_datapoints(self):
        self.assertIsNotNone(Spectrum(self.x, self.y))

    def test_creation_with_custom_curve(self):
        self.assertIsNotNone(Spectrum(self.x, self.y, curve=Curve()))

    def test_creation_with_custom_focus(self):
        self.assertIsNotNone(Spectrum(self.x, self.y, focus=LineSubset()))

    def test_creation_with_equal_sigma_type(self):
        self.assertIsNotNone(Spectrum(self.x, self.y, sigma_type='equal'))

    def test_creation_with_huber_sigma_type(self):
        self.assertIsNotNone(Spectrum(self.x, self.y, sigma_type='huber'))

    def test_length(self):
        self.assertEqual(len(Spectrum(self.x, self.y)), 3)

    def test_true_if_has_datapoints(self):
        self.assertTrue(Spectrum(self.x, self.y))

    def test_false_if_empty(self):
        self.assertFalse(Spectrum())

    def test_f(self):
        self.assertAlmostEqual(
            sum(Spectrum(self.x, self.y, curve=Curve(lambda x: x)).f), 6.0)

    def test_delta(self):
        self.assertAlmostEqual(
            sum(Spectrum(self.x, self.y, curve=Curve(lambda x: x)).delta), 4.2)

    def test_si_equal(self):
        self.assertAlmostEqual(
            sum(Spectrum(self.x, self.y, curve=Curve(lambda x: x)).si), 3.0)

    def test_si_huber(self):
        self.assertAlmostEqual(1.132992,
            sum(Spectrum(self.x, self.y, sigma_type='huber').si))

    def test_si_unknown(self):
        with self.assertRaises(ValueError):
            _ = Spectrum(self.x, self.y, sigma_type='nonexistent sigma type').si

    def test_mse(self):
        self.assertAlmostEqual(
            Spectrum(self.x, self.y, curve=Curve(lambda x: x)).mse, 2.92)

    def test_domain(self):
        self.assertEqual(Spectrum(self.x, self.y).domain, LineSubset(1.0, 3.0))

    def test_focused(self):
        self.assertAlmostEqual(4.6,
            sum(Spectrum(self.x, self.y, focus=LineSubset(0.5, 2.5)).focused.y))

    def test_within(self):
        self.assertAlmostEqual(
            sum(Spectrum(self.x, self.y).within(LineSubset(0.5, 2.5)).y), 4.6)

    def test_focus_on_edge(self):
        spectrum = Spectrum(self.x, self.y)
        spectrum.focus_on_edge(width=0.5)
        self.assertAlmostEqual(sum(spectrum.focused.y), 6.8)

    def test_focus_on_whole(self):
        spectrum = Spectrum(self.x, self.y)
        spectrum.focus_on_whole()
        self.assertAlmostEqual(sum(spectrum.focused.y), 10.2)

    def test_focus_on_points(self):
        spectrum = Spectrum(self.x, self.y)
        spectrum.focus_on_points(points=(2.0,), width=1.5)
        self.assertAlmostEqual(sum(spectrum.focused.y), 3.4)


if __name__ == '__main__':
    unittest.main()



import os
import tkinter.filedialog as fd
import tkinter.messagebox as mb


def open_file_dialogue():
    title = 'Open ruby spectrum file...'
    filetypes = (("Text files", "*.txt"), ("All files", "*.*"))
    wdir = os.getcwd()
    return fd.askopenfilename(title=title, filetypes=filetypes, initialdir=wdir)


def show_about():
    message = 'pRuby - pressure estimation based on ruby fluorescence ' \
              'spectrum by Daniel Tcho\u0144\n\n' \
              'For details and help, visit https://github.com/Baharis/pRuby ' \
              'or contact me on https://dtools.pl'
    mb.showinfo(title='About pRuby', message=message)



from typing import Callable, Iterable
import numpy as np


class Curve:
    def __init__(self, func: Callable = lambda x: 0, args=tuple()):
        self.func: Callable = func
        self.args: Iterable = args
        self.uncs: tuple = tuple(np.zeros_like(args))

    def __call__(self, *args):
        x, args = self._interpret_call(args)
        try:
            return np.array([self.func(x_val, *args) for x_val in x])
        except TypeError:
            return self.func(x, *args)

    def _interpret_call(self, args):
        call_x, call_args = args[0], args[1:]
        args = list(self.args)
        for i, a in enumerate(call_args):
            try:
                args[i] = a
            except IndexError as e:
                raise IndexError('Too many arguments provided to the function')
        return call_x, args
